import Foundation

/// A Linear Congruential `RandomNumberGenerator` of pseudo-random values
///
/// - Note: The LCRNG is independent. Values generated by one instance will have no effect on others
/// - Note: The LCRNG is deterministic. Any instance initialized by the same seed will generate the same sequence of values
///
/// - SeeAlso: https://nuclear.llnl.gov/CNP/rng/rngman/node4.html
/// - SeeAlso: https://en.wikipedia.org/wiki/Linear_congruential_generator
struct LinearCongruentialRandomNumberGenerator: RandomNumberGenerator {
    private var seed: UInt64

    /// Initializes a `LCRNG` with the provided seed
    ///
    /// Any two `LCRNG` instances initialized by the same seed will independently generate the same sequence of pseudo-random
    ///
    /// - Parameter seed: The seed value which should be used to start the generator
    init(seed: UInt64) {
        self.seed = seed
    }

    private let a: UInt64 = 2862933555777941757
    private let b: UInt64 = 3037000493
    mutating func next() -> UInt64 {
        seed = a &* seed &+ b
        return seed
    }
}
